
# Copyright (C) 2023-2025 Cognizant Digital Business, Evolutionary AI.
# All Rights Reserved.
# Issued under the Academic Public License.
#
# You can be released from the terms, and requirements of the Academic Public
# License by purchasing a commercial license.
# Purchase of a commercial license is mandatory for any use of the
# neuro-san SDK Software in commercial settings.
#
# END COPYRIGHT

from neuro_san.coded_tools.math_guy.calculator import Calculator as BaseCalculator


# No need to repeat code, we just want to be sure that the deep reference to the
# coded tool can be reached.
Calculator = BaseCalculator
